/********************************************************************************
Copyright (C) Takuya.S

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
********************************************************************************/
#include <stdio.h>
#include <signal.h>

#ifdef WIN32
#include "stdafx.h"
#endif

#include "Esperance.h"

bool* sig_stop = NULL;

static void sigIntHandler(int s) {
  *sig_stop = false;
}

#ifdef CUI
int main(int argc, char** argv) {
  
  Esperance* esp = NULL;
  
  char out_file[] = "output.mp4";
  
  if (argc < 2) {
    fprintf(stderr, "usage: Esperance inputfile.m2p\n");
    return EXIT_SUCCESS;
  }
  
  esp = new Esperance();
  
  sig_stop = esp->getCtrlBool();
  signal(SIGINT, sigIntHandler);
  
  //ESP:ACTION
  esp->init();  
  
  // allocate memory for each params
  esp->ffmpeg_param = (esp_ffmpeg_params_t*)malloc(sizeof(esp_ffmpeg_params_t));
  esp->x264_param = (esp_x264_params_t*)malloc(sizeof(esp_x264_params_t));
  esp->faac_param = (esp_faac_params_t*)malloc(sizeof(esp_faac_params_t));
  
  //set input file
  //set output file
  esp->ffmpeg_param->inputfile  = argv[1];
  esp->params->outputfileName = out_file;
  
  //ESP:ACTION
  if (esp->initDecodeCodec() == false) {
    esp->close();
    delete esp;
    return EXIT_FAILURE;
  }

  /* preset level
   * 0: normal
   * 1: high quality
   * 2: high quality & high profile
   */
  if (argc >= 5) {
    esp->params->h264HighProfile = true;
  } else {
    esp->params->h264HighProfile = false;
  }
  esp->params->x264_bframe = false;
  
  esp->presetParams(esp, argc-2);
  
  //ESP:ACTION
  esp->initCodecs();

  //ESP:ENCODING
  esp->startEncoding();

  //ESP:CLOSE & FREE
  esp->closeCodecs();
  esp->close();

  free(esp->ffmpeg_param);
  free(esp->x264_param);
  free(esp->faac_param);

  delete esp;

  return EXIT_SUCCESS;
  
}
#endif

bool Esperance::presetParams(Esperance* esp, int preset) {

  esp->params->autoDetectProgressive = true;
  
  esp->params->deinterlace = 1;
  if (esp->getInputWidth() < 600) 
    esp->params->deinterlace = 0;
  
  esp->params->inputWidth = esp->getInputWidth();
  esp->params->inputHeight = esp->getInputHeight();
  
  esp->params->outputWidth = esp->getInputWidth();
  esp->params->outputHeight = esp->getInputHeight();
  esp->params->needResize = false;
  esp->x264_param->resizedPicture = false;

  // set profile level
  esp->x264_param->transform_8x8 = esp->params->h264HighProfile;
  
  if (esp->params->h264HighProfile == true) {
    ;
  }else if (esp->getAspectX() == 32 && esp->getAspectY() == 27) {
    esp->params->outputWidth = 854;
    esp->params->outputHeight = 480;
    esp->params->needResize = true;
    
    esp->x264_param->resizedPicture = true;
    
  } else if (esp->getAspectX() == 8 && esp->getAspectY() == 9) {
    esp->params->outputWidth = 720;
    esp->params->outputHeight = 540;
    esp->params->needResize = true;
    
    esp->x264_param->resizedPicture = true;
  }
  
  // init faac_params
  if (preset == 2) {
    esp->faac_param->bitrate = 192;
  } else if (preset == 1) {
    esp->faac_param->bitrate = 160;
  } else {
    esp->faac_param->bitrate = 128;
  }
  
  esp->faac_param->tns     = 1;
  
  // init x264_params
  if (preset >= 2) {
    esp->x264_param->subpel_refine = 7;
    esp->x264_param->frame_reference = 3;
    esp->x264_param->crf = 16;
    esp->x264_param->qp_min = 10;
    esp->x264_param->qp_max = 20;
    esp->x264_param->bframe = (esp->params->x264_bframe == true) ? 1 : 0;
    esp->x264_param->bframe_adaptive = true;
  } else if (preset == 1) {
    esp->x264_param->subpel_refine = 6;
    esp->x264_param->frame_reference = 3;
    esp->x264_param->crf = 21;
    esp->x264_param->qp_min = 15;
    esp->x264_param->qp_max = 25;
    esp->x264_param->bframe = (esp->params->x264_bframe == true) ? 2 : 0;
    esp->x264_param->bframe_adaptive = true;
  } else {
    esp->x264_param->subpel_refine = 4;
    esp->x264_param->frame_reference = 2;
    esp->x264_param->crf = 26;
    esp->x264_param->qp_min = 16;
    esp->x264_param->qp_max =28;
    esp->x264_param->bframe = (esp->params->x264_bframe == true) ? 3 : 0;
    esp->x264_param->bframe_adaptive = true;
  }

  esp->x264_param->pass = 0;
  esp->x264_param->e_mode = X264_QMODE_CRF;
  esp->x264_param->qp_step = 3;
  esp->x264_param->me_method = 2;
  esp->x264_param->key_max = 30;
  esp->x264_param->threads = 2;

  esp->x264_param->df_alpha = 0;
  esp->x264_param->df_beta = 0;
  esp->x264_param->nr = 0;
  esp->x264_param->weighted_bipred = true;
  esp->x264_param->bidir_me = true;
  esp->x264_param->fast_pskip = true;

  return true;
}

Esperance::Esperance() {
  params = new EsperanceParams();
  
  params->deinterlace = 1;
  params->encodeVideo = true;
  params->encodeAudio = true;

  params->mp4_vSync = false;
  params->mp4_aSync = false;
  
  params->mp4_vDelay = 0;
  
  ctr  = true;
  stop = false;
  susp = false;
  
  videoEncodedFrameCount = 0;

  videoPTS = 0;
  audioPTS = 0;
  
  progStDuration = 3003;
  progVideoFrameCount = 0;
  progVideoTotalDuration = 0;

  params->colorCorrection = true;
  
}

Esperance::~Esperance() {
  delete params;
}

bool Esperance::init() {

  ffmpeg  = new FFMPEGDecoder();
  x264 = new X264Encoder();
  faac = new FAACEncoder();
  mp4 = new MP4Handler();

  return true;
}

bool Esperance::initDecodeCodec() {

  fprintf(stderr, "esp: init decode codec\n");

  if (ffmpeg->open(ffmpeg_param) == false) {
    return false;
  }

  ffmpeg->init();
  
  return true;
}

bool Esperance::initCodecs() {
  
  ffmpeg->setVideoResampleContext(params->outputWidth,
                                  params->outputHeight);
  
  faac_param->samplerate  = ffmpeg->getAudioSampleRate();
  faac_param->numChannels = ffmpeg->getAudioChannels();
                 
  faac->init(faac_param);                
  faac->open();

  x264_param->width   = params->outputWidth;
  x264_param->height  = params->outputHeight;
  
  x264_param->sar_width  = ffmpeg->getAspectX();
  x264_param->sar_height = ffmpeg->getAspectY(); 
  
  x264_param->fps_den = ffmpeg->getFpsDen();
  x264_param->fps_num = ffmpeg->getFpsNum();
  
#ifndef WIN32
  char* x264_state_log =  (char* )malloc(512 + 64);
  char x264_state_log2[512];
  const char fname[] = "/x264_1st_pass_log";
  //getcwd(x264_state_log2,sizeof(x264_state_log2));
  int l = strlen(getenv("HOME"));
  strncpy(x264_state_log2,getenv("HOME"), l);
  x264_state_log2[l] = '\0';
  //strcat(x264_state_log2,getenv("HOME"));
  fprintf(stderr, "file name : %s\n", x264_state_log2);
  strcat(x264_state_log2, fname);
  fprintf(stderr, "file name : %s\n", x264_state_log2);
  sprintf(x264_state_log, "%s", x264_state_log2);
  x264_param->log_file = x264_state_log;
#endif

  x264->init(x264_param);
  x264->open();

  mp4->init();  
  mp4->open(params->outputfileName);
  mp4->setMP4TimeScale(90000);  
  mp4->setVideoDelayTime(params->mp4_vDelay);

  if (ffmpeg->hasVideo() == true && params->encodeVideo == true) {
    mp4->addH264VideoTrack(x264->getSPSRef(), params->outputWidth, params->outputHeight, 90000);
    mp4->setH264SequenceParameter(x264->getSPSRef(), x264->getSPSLen());
    mp4->setH264PictureParameter (x264->getPPSRef(), x264->getPPSLen());
  }

  if (ffmpeg->hasAudio() == true && params->encodeAudio == true) {
    faac->getDecoderSpecificInfo(mp4->getASCRef(), mp4->getASCLengthRef());
    mp4->addAACAudioTrack(faac->getSampleRate());
  }
  
  return true;
}

bool Esperance::setParams() {
  return true;
}

bool Esperance::startEncoding() {

  time(&startTime);

  if (x264_param->pass == 2 && x264_param->e_mode == X264_QMODE_CBR ) {//&& params->inputs == 1) {
    
    encodeX264FirstPass();
    
    if (ctr == false) {
      return false;
    }
    
    ffmpeg->close();
    delete ffmpeg;
    ffmpeg = new FFMPEGDecoder();
    
    ffmpeg->open(ffmpeg_param);
    ffmpeg->init();
    
    ffmpeg->setVideoResampleContext(params->outputWidth,
                                    params->outputHeight);
    
    x264_param->pass = 1;
    
    x264->close();
    delete x264;
    
    x264 = new X264Encoder();
    
    x264->init(x264_param);
    x264->open();
    
    videoEncodedFrameCount = 0;
  }
  
  encodeX264SecondPass();
  
  for (int i = 1; i < params->inputs && x264_param->pass != 2 && x264_param->e_mode != X264_QMODE_CBR; i++) {
    
    ffmpeg->close();
    delete ffmpeg;
    ffmpeg = new FFMPEGDecoder();
    
    ffmpeg->open((char*)params->inputfiles[i]);
    ffmpeg->init();
    
    ffmpeg->setVideoResampleContext(params->outputWidth,
                                    params->outputHeight);
    
    params->needResize = true;
    
    encodeX264SecondPass();
  }
  
  flush();

  fprintf(stderr, "sys: done - video pts %lld\n", videoPTS);
  
  fprintf(stderr, "sys: FPS %.2lf\n", getFPS());
  
  return true;
}

double Esperance::getFPS()
{
  int64_t dvf = ffmpeg->getDecodedVideoFrameCNT();
  double dt = 0.0;

  if (dvf == 0) {
    return 0.0;
  } else {
    
    time(&endTime);
    dt = difftime(endTime, startTime);
    
  }
  
  return (double)(dvf) / dt;
}

bool Esperance::encodeX264FirstPass() {
  int lim = 0;
  
  pic = x264->getX264PictureRef();
  
  //ctr: will be changed by NSThread for INT
  for(; ctr == true; lim++) {
    
    if (susp == true) {
      
#ifndef WIN32
      sleep(5);
#endif
      
      fprintf(stderr, "sys: suspending\n");
      continue;
    }
    
    ///// if EOF stop encoding
    if (ffmpeg->isFinished() == true) {
      fprintf(stderr, "dec: EOF done!!\n");
      return true;
    }
    /////
    
    /////
    
    ffmpeg->decode();
    
    /////
    if (ffmpeg->isDoneVideo() != 0 && params->encodeVideo == true) {
      fprintf(stderr, "sys: 1st-pass video processing %d\n", ++videoEncodedFrameCount);
      
      AVPicture* tmpP = NULL;
      
      if (params->deinterlace == 0) {
        tmpP = ffmpeg->getVideoYUVRef();
      } else {
        tmpP = ffmpeg->getDeinterlacedVideoYUVRef();
      }
      
      if (params->needResize == true) {
        copy_frame(ffmpeg->getResampledVideoYUVRef(tmpP), pic);
      } else {
        copy_frame(tmpP, pic);
      }
      
      pic->i_pts = ffmpeg->getPts();
      
      x264->encode();
      
    } 
  }
  
  if (ctr == false) {
    fprintf(stderr, "sys: interrupted\n");
  }
  
  return true;
}


bool Esperance::encodeX264SecondPass() {
  int lim = 0;
  int i = 0;
  int k = 0;
  int f = 0;
  
  initialI = true;
  
  pic = x264->getX264PictureRef();
  
  //ctr: will be changed by NSThread for INT
  for(i = 0; ctr == true; lim++) {
    
    if (susp == true) {
      fprintf(stderr, "sys: suspending\n");
      
#ifndef WIN32             
      sleep(5);
#endif
      
      continue;
    }
    
    ///// if EOF stop encoding
    if (ffmpeg->isFinished() == true) {
      fprintf(stderr, "ffmpeg: EOF done!!\n");
      return true;
    }
    /////
    
    /////
    if (faac->isNeedBuffer() == false && params->encodeAudio == true) {
      
      faac->encode();
      
      if (faac->getBytesWritten() > 0) {
        fprintf(stderr, "sys: audio processing %d %d\n", ++k, ffmpeg->getDecodedAudioSampleNum());
        mp4->writeMP4AudioSample(faac->getEncodedBufferRef(),
                                 faac->getEncodedBytes(),
                                 faac->getEncodedSamples() / faac->getNumChannels());
        fprintf(stderr, "\n");
      }
      continue;
    }
    /////
    
    ffmpeg->decode();
    
    /////
    if (ffmpeg->isDoneVideo() != 0 && params->encodeVideo == true) {
      fprintf(stderr, "sys: video processing %d\n", ++videoEncodedFrameCount);
      
      AVPicture* tmpP = NULL;
      AVPicture* rgbP = NULL;
      
      if (params->deinterlace == 1 && ffmpeg->isInterlaced() == true) {
        tmpP = ffmpeg->getDeinterlacedVideoYUVRef();
      } else {
        tmpP = ffmpeg->getVideoYUVRef();                
      }
      
      // color correction: Y & S up
      if (params->colorCorrection == true) {
        rgbP = ffmpeg->convertYUVtoRGB(tmpP);
        correctColor(rgbP->data[0]);
        ffmpeg->convertRGBtoYUV(rgbP, tmpP);
      }
      
      if (params->needResize == true) {
        copy_frame(ffmpeg->getResampledVideoYUVRef(tmpP), pic);
      } else {
        copy_frame(tmpP, pic);
      }
      
      //////////
      // YUV420 to H.264      
      pic->i_pts = ffmpeg->getPts();
      
      x264->encode();
      /////////
      
      if ( x264->getBytesWritten() > 0) {
        
        if (params->mp4_vHack == true) {
          f = 1;
        } else {
          f = ((x264->getSliceType() - 1) == 0)? 1 : 0;
        }
        
        if (params->x264_bframe == true) {
          int slice =  x264->getSliceType();
          
          if (slice == 1 || slice == 2) { // I
            handleI();
          } else if (slice == 3) { // P
            handleP();
          } else if (slice == 4 || slice == 5) { // B
            handleB();
          }
          
        } else {
          
          mp4->writeMP4VideoSample(x264->getNalEncodedBufferRef(),
                                   x264->getTotalBytesWritten(),
                                   ffmpeg->getNextVideoDuration(),
                                   f);
        }
      }
      
      fprintf(stderr, "\n");
    } else if (ffmpeg->isDoneAudio() != 0 && params->encodeAudio == true) { // AUDIO
      
      fprintf(stderr, "sys: audio processing %d %d\n", ++k, ffmpeg->getDecodedAudioSampleNum());
      
      faac->encode((int32_t *)ffmpeg->getAudioRAWRef(),
                   (int32_t  )ffmpeg->getDecodedAudioSampleNum() / 2);
      
      if (faac->getBytesWritten() > 0) {
        mp4->writeMP4AudioSample(faac->getEncodedBufferRef(),
                                 faac->getEncodedBytes(),
                                 faac->getEncodedSamples() / faac->getNumChannels());
        fprintf(stderr, "\n");
      }
    }
    /////
  }
  
  if (ctr == false) {
    fprintf(stderr, "sys: interrupted\n");
  }
  
  return true;
}


bool Esperance::flush() {
  
  int i = 0;
  int f;
  
  do {
    
    if (ffmpeg->hasVideo() == false) {
      break;
    }
    
    x264->flush();
    
    if ( x264->getBytesWritten() > 0 ) {
      
      fprintf(stderr, "esp: flushing video\n");
      
      if (params->mp4_vHack == true) {
        f = 1;
      } else {
        f = ((x264->getSliceType() - 1) == 0)? 1 : 0;
      }
      
      if (params->x264_bframe == true) {
        int slice =  x264->getSliceType();
        
        if (slice == 1 || slice == 2) { // I
          handleI();
        } else if (slice == 3) { // P
          handleP();
        } else if (slice == 4 || slice == 5) { // B
          handleB();
        }
        
      } else {
        mp4->writeMP4VideoSample(x264->getNalEncodedBufferRef(),
                                 x264->getTotalBytesWritten(),
                                 ffmpeg->getNextVideoDuration(),
                                 f);
      }
      
    }
  } while(x264->getBytesWritten() > 0  && ++i < 10);
  
  if (ffmpeg->hasVideo() == true && params->x264_bframe == true && pSize != 0) {
    
    fprintf(stderr, "esp: flushing last P-Frame\n");
    
    mp4->writeMP4VideoSample(pBuf,
                             pSize,
                             pDuration,
                             3003,
                             0);
  }
  
  if (params->encodeAudio == true) {
    
    faac->flush();
    
    if (faac->getBytesWritten() > 0) {
      
      fprintf(stderr, "esp: flushing audio\n");
      
      mp4->writeMP4AudioSample(faac->getEncodedBufferRef(),
                               faac->getEncodedBytes(),
                               faac->getEncodedSamples() / faac->getNumChannels());
      fprintf(stderr, "\n");
    }
  }
  
  return true;
}

bool Esperance::stopEncoding() {
  
  ctr = false;
  stop = true;
  
  return true;
}

bool Esperance::suspendEncoding() {
  if (susp == true) {
    susp = false;
  } else {
    susp = true;
  }
  return true;
}

bool* Esperance::getCtrlBool() {
  return &ctr;
}

bool Esperance::closeCodecs() {
  
  mp4->close();
  x264->close();
  ffmpeg->close();
  faac->close();
  
  return true;
}

bool Esperance::close() {
  
  delete x264;
  delete faac;
  delete ffmpeg;
  delete mp4;
  
  return true;
}


void Esperance::copy_frame(AVPicture *pin, x264_picture_t  *pout) {

  //pin = ffmpeg->getVideoYUVRef();             

  // slow but stable
  int i, x, y;
  int stride0 = pin->linesize[0];
  int stride1 = pin->linesize[1];
  int stride2 = pin->linesize[2];
  
  int Y;
  
  for (y = 0, i = 0; y < params->outputHeight; y++) {
    for (x = 0; x < params->outputWidth; x++) {
      Y = pin->data[0][stride0 * y + x];// - 15; // Y
      Y = (Y > 255)?255:Y;
      
      //pout->img.plane[0][i++] = pin->data[0][stride0 * y + x]; // Y
      pout->img.plane[0][i++] = Y;
    }
  }
  
  for (y = 0, i = 0; y < params->outputHeight / 2; y++) {
    for (x = 0; x < params->outputWidth / 2; x++) {
      pout->img.plane[1][i]   = pin->data[1][stride1 * y + x]; // U
      pout->img.plane[2][i++]   = pin->data[2][stride2 * y + x]; // U
    }
  }
  
}                            

int Esperance::getInputWidth() {
  return ffmpeg->getWidth();
}

int Esperance::getOutputWidth() {
  return ffmpeg->getWidth();
}

int Esperance::getInputHeight() {
  return ffmpeg->getHeight();
}

int Esperance::getOutputHeight() {
  return ffmpeg->getHeight();
}

int Esperance::getInputAudioSamplerate() {
  return ffmpeg->getAudioSampleRate();
}

int Esperance::getInputAudioNumChannel() {
  return ffmpeg->getAudioChannels();
}

int Esperance::getFpsDen() {
  return ffmpeg->getFpsDen();
}

int Esperance::getFpsNum() {
  return ffmpeg->getFpsNum();
}

int Esperance::getAspectX() {
  return ffmpeg->getAspectX();
}

int Esperance::getAspectY() {
  return ffmpeg->getAspectY();
}

int Esperance::getVideoEncodedFrameCount() {
  return videoEncodedFrameCount;
}

int64_t Esperance::getDecodedBytes() {
  return ffmpeg->getBytePosition();
}

int64_t Esperance::getInputFileSize() {
 return ffmpeg->getFileSize();
}

int Esperance::getProgress() {
  return ffmpeg->getProgress();
}

// handling each frame type for b support
void Esperance::handleI() {

  int offset = 0;
  int dul[8];
  
  if (pSize > 0) {
    
    if (bFrames == 0) {
      offset = pDuration;
    } else {
      for (int i = 0; i < bFrames; i++) {
	dul[i] = bDuration[i];
        offset += dul[i];
      }

      offset += pDuration;
    }
    
    mp4->writeMP4VideoSample(
                             pBuf,
                             pSize,
                             pDuration,
                             offset,
                             0);
    pSize = 0;
  }
  
  if (bFrames > 0) {
    int p = 0;
    
    for (int ii = 0; ii < bFrames; ii++) {
      
      mp4->writeMP4VideoSample(bBuf + p,
                               bSize[ii],
                               dul[ii]);
      p += bSize[ii];
    }
    bFrames = 0;
    bBufInd = 0;
  }
  
  if (initialI == true) { // avoid initial white I frame
    mp4->writeMP4VideoSample(x264->getNalEncodedBufferRef(),
                             x264->getTotalBytesWritten(),
                             0,
                             0,
                             1);
    initialI = false;
  }

  int d = ffmpeg->getNextVideoDuration();
  
  mp4->writeMP4VideoSample(x264->getNalEncodedBufferRef(),
                           x264->getTotalBytesWritten(),
                           d,
                           d,
                           1);
}


/*
 */
void Esperance::handleP() {
  int offset = 0;
  int dul[8];
  
  if (pSize > 0) {
    
    if (bFrames == 0) {
      offset = pDuration;
    } else {
      
      for (int i = 0; i < bFrames; i++) {
	dul[i] = bDuration[i];
        offset += dul[i];
      }
      offset += pDuration;
    }
    
    mp4->writeMP4VideoSample(pBuf,
                             pSize,
                             pDuration,
                             offset,
                             0);
    pSize = 0;
  }
  
  if (bFrames > 0) {
    int p = 0;
    
    for (int ii = 0; ii < bFrames; ii++) {
      mp4->writeMP4VideoSample(bBuf + p,
                               bSize[ii],
                               dul[ii]);
      
      p += bSize[ii];
    }
    bFrames = 0;
    bBufInd = 0;
  }
  
  storeP(x264->getNalEncodedBufferRef(),
         x264->getTotalBytesWritten(),
         ffmpeg->getNextVideoDuration());
}


void Esperance::handleB() {
  storeB(x264->getNalEncodedBufferRef(),
         x264->getTotalBytesWritten(),
         ffmpeg->getNextVideoDuration());
}


void Esperance::storeP(u_int8_t* data, int size, int duration) {
  pDuration = duration;
  pSize = size;
  memcpy(pBuf, data, size);
}


void Esperance::storeB(u_int8_t* data, int size, int duration) {
  bSize[bFrames] = size;
  bDuration[bFrames] = duration;
  memcpy(bBuf + bBufInd, data, size);
  
  bBufInd += size;
  bFrames++;
}


/////////////////////////////////////////////////
void Esperance::correctColor(u_int8_t *pic) {

  int ih;
  int is;
  int ib;

  int r, g, b;
  int max, min, mami;

  for (int i = 0; i < params->inputWidth * params->inputHeight *3; i+= 3) {
    
    r = pic[i  ];
    g = pic[i+1];
    b = pic[i+2];

    max = (r > g) ? r : g;
    max = (max > b) ? max :b;

    min = (r < g) ? r : g;
    min = (min < b) ? min : b;

    mami = max - min;

    // H
    if (mami == 0) {
      ih = 0;
    } else {

      if (r == max) {
        
        if (b == min) {
          ih = (g - min) * 60 / mami; // RGB
        } else {
          ih = (max - b) * 60 / mami + 300; // RBG
        }
        
      } else if (g == max) {
        
        if (b == min) {
          ih = (max - r) * 60 / mami + 60;
        } else {
          ih = (b - min) * 60 / mami + 120;
        }

      } else {
        
        if (r == min) {
          ih = (max - g) * 60 / mami + 180;
        } else {
          ih = (r - min) * 60 / mami + 240;
        }
        
      }
    }

    // S
    if (max != 0) {
      is = mami * 255 / max;
    } else {
      is = 0;
    }
    
    is = is + 7;
    is = ( is > 255 )? 255 : is;
    
    mami = is * max / 255;
    
    // B
    ib = max + 0;

    ///////////
    if (ih < 60) {
      r = max;
      g = min + ih*mami/60;
      b = min;
    } else if (ih < 120) {
      r = max - (ih-60)*mami/60;
      g = max;
      b = min;
    } else if (ih < 180) {
      r = min;
      g = max;
      b = min + (ih-120)*mami/60;
    } else if (ih < 240) {
      r = min;
      g = max - (ih-180)*mami/60;
      b = max;
    } else if (ih < 300) {
      r = min + (ih-240)*mami/60;
      g = min;
      b = max;
    } else {
      r = max;
      g = min;
      b = max - (ih-300)*mami/60;
    }

    r = (r > 255)?255:r;
    g = (g > 255)?255:g;
    b = (b > 255)?255:b;

    pic[i  ] = r;
    pic[i+1] = g;
    pic[i+2] = b;
  }
}
