/********************************************************************************
Copyright (C) Takuya.S

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "FAACEncoder.h"

/**********************************************************************
CONSTRUCTOR & DESTRUCTOR
**********************************************************************/

////////////////////////////////////////
FAACEncoder::FAACEncoder() {
  _param = NULL;
  
  _totalBytesWritten = 0;
  _internalBufferRemain = 0;

  //_asc =  (u_int8_t *)malloc(256);

  _encodedBuffer = (u_int8_t *)malloc(8192);
  _internalBuffer = (int32_t *)malloc(8192 * sizeof(int32_t));

  _ahureBuffer = (int32_t *)malloc(4096 * sizeof(int32_t));
  _audioFrame = (int32_t *)malloc(4096 * (sizeof(int32_t)));
}

////////////////////////////////////////
FAACEncoder::FAACEncoder(unsigned long sampleRate, int numChannels, u_int32_t bitRate) {
  
  _param = NULL;
  
  _sampleRate  = sampleRate;
  _numChannels = numChannels;

  _bitRate = bitRate * 1000;

  _totalBytesWritten = 0;
  _internalBufferRemain = 0;

  _encodedBuffer = (u_int8_t *)malloc(8192);
  _internalBuffer = (int32_t *)malloc(8192 * sizeof(int32_t));

  _ahureBuffer = (int32_t *)malloc(4096 * sizeof(int32_t));
  _audioFrame = (int32_t *)malloc(4096 * (sizeof(int32_t)));
}

////////////////////////////////////////
FAACEncoder::~FAACEncoder() {

  if (_encodedBuffer != NULL) {
    free(_encodedBuffer);
    _encodedBuffer = NULL;
  }

  if (_internalBuffer != NULL) {
    free(_internalBuffer);
    _internalBuffer = NULL;
  }

  if (_ahureBuffer != NULL) {
    free(_ahureBuffer);
    _ahureBuffer = NULL;
  }

  if (_audioFrame != NULL) {
    free(_audioFrame);
    _audioFrame = NULL;
  }

  fprintf(stderr, "faac: object deleted\n");
}

/**********************************************************************
PUBLIC METHOD
**********************************************************************/

////////////////////////////////////////
bool FAACEncoder::init() {

  _bitRate = 128 * 1000;

  if (_encodedBuffer == NULL) {
    fprintf(stderr, "faac: malloc error\n");
    return false;
  }
  return true;
}

////////////////////////////////////////
bool FAACEncoder::init(int samplerate, int channels) {

  _bitRate = 128 * 1000;
  
  _sampleRate  = samplerate;
  _numChannels = channels;

  return true;
}

////////////////////////////////////////
bool FAACEncoder::init(esp_faac_params_t* p) {
  _param = p;
  
  _sampleRate  = _param->samplerate;
  _numChannels = _param->numChannels;
  
  _bitRate     = _param->bitrate * 1000;

  return true;
}

////////////////////////////////////////
void FAACEncoder::setDefaultEncoderCfg() {

  _cfgEncoder->mpegVersion   = MPEG4;
  _cfgEncoder->aacObjectType = LOW;
  _cfgEncoder->allowMidside  = 0; // Don't touch this parameter
  _cfgEncoder->useLfe        = 0;
  _cfgEncoder->useTns        = 1; // Use Temporal Noise Shaping
  _cfgEncoder->bitRate       = _bitRate / _numChannels;
  _cfgEncoder->bandWidth     = 0; // FIXME
  _cfgEncoder->quantqual     = 10000; // FIXME
  _cfgEncoder->outputFormat  = 0;
  //_cfgEncoder->inputFormat   = FAAC_INPUT_FLOAT;
  _cfgEncoder->inputFormat   = FAAC_INPUT_16BIT;
  _cfgEncoder->shortctl      = SHORTCTL_NORMAL;
  //_cfgEncoder->shortctl      = SHORTCTL_NOSHORT;
}

// open faac codec
////////////////////////////////////////
bool FAACEncoder::open() {

  if (!(_hEncoder = faacEncOpen(_sampleRate,
                                _numChannels,
                                &_maxInputSamples,
                                &_maxOutputBytes))) {
    return false;
  }

  _maxInputSamples2 = _maxInputSamples / _numChannels;

  if (!(_cfgEncoder = faacEncGetCurrentConfiguration(_hEncoder))) {
    return false;
  } 
  
  if (_param == NULL) {
        setDefaultEncoderCfg();
  } else {
    setDefaultEncoderCfg();
        
    _cfgEncoder->bitRate = _bitRate / _numChannels;
    _cfgEncoder->useTns  = _param->tns; // Use Temporal Noise Shaping
  }
  
  faacEncSetConfiguration(_hEncoder, _cfgEncoder);
  
  fprintf(stderr, "faac: codec opened\n");
  return true;
}

// close faac codec
////////////////////////////////////////
bool FAACEncoder::close() {
  if (_hEncoder == NULL) {
    return false;
  }

  if (faacEncClose(_hEncoder) == EXIT_SUCCESS) {
    fprintf(stderr, "faac: codec closed\n");
    return true;
  } else {
    return false;
  }
}


////////////////////////////////////////
bool FAACEncoder::encode() {

  _bytesWritten = 0;
  int32_t *audio = NULL;
  
  //audio = getFIFOBuffer();
  audio = getFIFOBuf();
  
  _bytesWritten = faacEncEncode(_hEncoder,
                                audio,
                                _maxInputSamples,
                                _encodedBuffer,
                                _maxInputSamples);
  
  return true;
}

////////////////////////////////////////
bool FAACEncoder::flush() {

  _encodedSamples = _internalBufferRemain * _numChannels;

  _bytesWritten = faacEncEncode(_hEncoder,
                                _internalBuffer,
                                _encodedSamples,
                                _encodedBuffer,
                                _maxInputSamples);
  
  _internalBufferRemain = 0;
  
#ifdef DEBUG_FAAC_ENCODER
  fprintf(stderr,"faac: new encoded %d bytes\n", _bytesWritten);
  fprintf(stderr,"faac: encoded %d bytes\n", _bytesWritten);
  fprintf(stderr, "faac: total %lld bytes written\n\n", _totalBytesWritten);
#endif
  
  if (_bytesWritten <= 0) {
    return false;
  }

  return true;
}


////////////////////////////////////////
bool FAACEncoder::encode_old(int32_t *input, int32_t samples) {

  int32_t *audio = input;
  
  //fprintf(stderr, "faac: samples %d\n", samples);

  if (samples > (int32_t)_maxInputSamples) {
    putFIFOBuffer(input, samples);
    audio = getFIFOBuffer();
    manageFIFOBuffer();
    _encodedSamples = _maxInputSamples;
    //fprintf(stderr, "faac: type1 %d\n", _maxInputSamples);
  } else {
    audio = input;
    _encodedSamples = samples;
    //fprintf(stderr, "faac: type2 %d\n", samples);
  }
  
  _bytesWritten = faacEncEncode(_hEncoder,
                                audio,
                                _encodedSamples,
                                _encodedBuffer,
                                _maxInputSamples);

  if (_bytesWritten <= 0) {
    _encodedSamples = 0;
    return false;
  }
  return true;
}

////////////////////////////////////////
bool FAACEncoder::encode(int32_t *input, int32_t samples) {
  
  int32_t *audio = NULL;
  
  _bytesWritten = 0;
  
  // put fifo
  putFIFOBuf(input, samples);
  
  // get from fifo
  if ((audio = getFIFOBuf()) == NULL) {
    _encodedSamples = 0;
    return false;
  } else {
    _encodedSamples = _maxInputSamples;
  }
  
  // if data gained  yes -> encode: no -> skip
  _bytesWritten = faacEncEncode(_hEncoder,
                                audio,
                                _maxInputSamples,
                                _encodedBuffer,
                                _maxInputSamples);
  
  return true;
}


/************************************************************
 * GET
 ************************************************************/

////////////////////////////////////////
void FAACEncoder::getDecoderSpecificInfo(unsigned char **asc, long unsigned int *ascLength) {
  faacEncGetDecoderSpecificInfo(_hEncoder,
                                asc,
                                ascLength);
}

////////////////////////////////////////
void FAACEncoder::setDecoderSpecificInfo() {
  faacEncGetDecoderSpecificInfo(_hEncoder,
                                &_asc,
                                &_ascLength);
}

////////////////////////////////////////
u_int8_t *FAACEncoder::getASCRef() {
  return _asc;
}


////////////////////////////////////////
long unsigned int FAACEncoder::getASCLength() {
  return _ascLength;
}


////////////////////////////////////////  
u_int8_t *FAACEncoder::getEncodedBufferRef() {
  return _encodedBuffer;
}

////////////////////////////////////////  
u_int32_t FAACEncoder::getEncodedBytes() {
  return _bytesWritten;
}

////////////////////////////////////////  
int32_t FAACEncoder::getEncodedSamples() {
  return _encodedSamples;
}

////////////////////////////////////////  
u_int64_t FAACEncoder::getTotalEncodedBytes() {
  return _totalBytesWritten;
}

////////////////////////////////////////  
unsigned int FAACEncoder::getMaxOutputBytes() {
  return _maxOutputBytes;
}

////////////////////////////////////////  
u_int32_t FAACEncoder::getBytesWritten() {
  return _bytesWritten;
}

////////////////////////////////////////  
int32_t FAACEncoder::getNumChannels() {
  return _numChannels;
}

////////////////////////////////////////  
u_int32_t FAACEncoder::getSampleRate() {
  return _sampleRate;
}

////////////////////////////////////////  
bool FAACEncoder::isNeedBuffer() {
  if (_internalBufferRemain < 1024) {
    return true;
  } else {
    return false;
  }
}

/**********************************************************************
 * PRIVATE METHOD
 **********************************************************************/

////////////////////////////////////////
// FIFO BUFFER OPERATIONS             //
////////////////////////////////////////

/* * */
/* new fifo mechanism for LPCM. also works mp2, ac3 ;) */
/* * */

////////////////////////////////////////
void FAACEncoder::putFIFOBuf(int32_t *input, int32_t samples) {
        
  memcpy(_internalBuffer+_internalBufferRemain,
         input,
         samples * sizeof(int32_t) / 2);
  
  _internalBufferRemain += samples/2;
}


////////////////////////////////////////  
int32_t *FAACEncoder::getFIFOBuf(void) {
  
  if (_internalBufferRemain < _maxInputSamples2) {
    return NULL;
  }
  
  memcpy(_audioFrame,
         _internalBuffer,
         _maxInputSamples * sizeof(int32_t));
  
  _internalBufferRemain -= _maxInputSamples2;
  
  memcpy(_internalBuffer,
         _internalBuffer + _maxInputSamples2,
         _internalBufferRemain * sizeof(int32_t));
  
  return _audioFrame;
}


/* */
/* old fifo mechanism. will be delete near future. */
/* */
////////////////////////////////////////
void FAACEncoder::putFIFOBuffer(int32_t *input, int32_t samples) {
  int i;

  //fprintf(stderr, "internal buffer: %d\n", _internalBufferRemain);

  _fifoDIF = (samples - _maxInputSamples) / 2;
  
  if (_fifoDIF < 0) {
          _fifoDIF = 0;
  }

  for (i = 0; i < (int32_t)_maxInputSamples2; i++) {
    _internalBuffer[i + _internalBufferRemain] = input[i];
  }
  _internalBufferRemain += _maxInputSamples2;

  for (i = 0; i < _fifoDIF; i++) {
    _ahureBuffer[i] = input[i + _maxInputSamples2];
  }

}

////////////////////////////////////////  
int32_t *FAACEncoder::getFIFOBuffer(void) {
  
  memcpy(_audioFrame,
         _internalBuffer,
         _maxInputSamples * sizeof(int32_t));
  _internalBufferRemain -= _maxInputSamples2;
  memcpy(_internalBuffer,
         _internalBuffer + _maxInputSamples2,
         _internalBufferRemain * sizeof(int32_t));
  
  return _audioFrame;
}

////////////////////////////////////////
void FAACEncoder::manageFIFOBuffer(void) {
  
  memcpy(_internalBuffer+_internalBufferRemain,
         _ahureBuffer,
         _fifoDIF * sizeof(int32_t));
  _internalBufferRemain += _fifoDIF;
}

/**********************************************************************
TEST FUNCTION
**********************************************************************/
#ifdef FAAC_TEST
int main() {
  FaacEncoder obj;
  obj.init();
  obj.open();
  obj.encode(container);
  obj.close();
}
#endif
