/********************************************************************************
Copyright (C) Takuya.S

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
********************************************************************************/

#ifndef _FAAC_ENCODER_H
#define _FAAC_ENCODER_H

#ifdef WIN32
#include "MP4Handler.h"
#endif

#ifdef _USE_MAKE_
#include <faac.h>
#else
#include "faac.h"
#endif

typedef struct {
	
	int samplerate;
	int numChannels;

	int bitrate;
	
	int tns;
} esp_faac_params_t;

class FAACEncoder {
 private:
  faacEncHandle         _hEncoder;
  faacEncConfiguration* _cfgEncoder;

  int32_t _encodedSamples;
  u_int32_t _bytesWritten;
  u_int64_t _totalBytesWritten;

  unsigned long _sampleRate;
  unsigned int  _numChannels;
  u_int32_t     _bitRate;
  unsigned long _maxInputSamples ;  // setted by faacEncOpen()
  unsigned long _maxInputSamples2;  // setted by faacEncOpen()
  unsigned long _maxOutputBytes;    // setted by faacEncOpen()

  u_int8_t *_asc;
  long unsigned int _ascLength;

  u_int32_t _encodedBufferSize;
  u_int8_t  *_encodedBuffer;

  int _fifoDIF;
  int32_t *_audioFrame;
  int32_t *_internalBuffer;
  uint32_t _internalBufferRemain;
  int32_t *_ahureBuffer;

  void setDefaultEncoderCfg();
  void putFIFOBuffer(int32_t *, int32_t);
  int32_t *getFIFOBuffer(void);
  void manageFIFOBuffer(void);
  
  void putFIFOBuf(int32_t *, int32_t);
  int32_t *getFIFOBuf(void);
  void manageFIFOBuf(void);
  
  esp_faac_params_t *_param;
  
 public:
  FAACEncoder();
  FAACEncoder(unsigned long, int, u_int32_t);
  ~FAACEncoder();
  
  bool init();
  bool init(int, int);
  bool init(esp_faac_params_t* );
  
  bool open();
  bool setEncoder();
  bool encode();
  bool encode(int32_t *, int32_t);
  bool encode_old(int32_t *, int32_t);
  bool flush();
  bool close();
  
  bool setInputSampleRate(unsigned long);
  bool setInputNumChannels(unsigned long);
  bool isNeedBuffer();
  
  void setDecoderSpecificInfo();
  void getDecoderSpecificInfo(unsigned char **, long unsigned int *);

  u_int8_t *getASCRef();
  long unsigned int getASCLength();
  
  int32_t getNumChannels();
  u_int32_t getSampleRate();
  unsigned int getMaxOutputBytes();
  u_int32_t getBytesWritten();

  u_int8_t *getEncodedBufferRef();
  u_int32_t getEncodedBytes();
  int32_t getEncodedSamples();
  u_int64_t getTotalEncodedBytes();
  
};

#endif
