/********************************************************************************
Copyright (C) Takuya.S

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "FFMPEGDecoder.h"

/**********************************************************************
 *CONST & DEST
 **********************************************************************/

////////////////////////////////////////
FFMPEGDecoder::FFMPEGDecoder() {

  _params = NULL;
  _totalDecodedBytes = 0;
  
  //_avFormatCtx  = NULL;
  _avFormatCtx = av_alloc_format_context();
  _avFormatPrm = (AVFormatParameters*)malloc(sizeof(AVFormatParameters));

  _vCodecCtx    = NULL;
  _vCodec       = NULL;
  _vFrame       = NULL;
  _vPacket.data = NULL;
  _remainVideoBytes = 0;

  _vEncCodecCtx   = NULL;
  _vResampleCtx   = NULL;

  _aCodecCtx    = NULL;
  _aCodec       = NULL;
  _aFrame       = (int32_t *)malloc(sizeof(int16_t) * 16384);
  _aResampleFrame = (int16_t *)malloc(sizeof(int16_t) * 16384);

  _rawAudioBuffer = (uint8_t *)malloc(sizeof(uint8_t) * 16384);
  _audioMemory = _rawAudioBuffer;
  _remainAudioBytes = 0;

  _videoStream = CODEC_TYPE_UNKNOWN;
  _audioStream = CODEC_TYPE_UNKNOWN;

  _inputfileOpened = false;

  _hasVideo = false;
  _hasAudio = false;

  _finished = false;
  _finishedReading = false;

  _packet.data = NULL;

  _videoDecodedFrameCNT = 0;
  
  _preVideoFrameDuration = 90000;
  _changedVideoFrameDuration = false;
  
  _dRingSize = 32;
  _dRingData = 0;
  _dRingMemory = 0;

  _pts = 0;
  _dts = 0;
  _dtsRingSize = 32;
  _dtsRingData = 0;
  _dtsRingMemory = 0;

  _videoDTS = 0;
  _audioDTS = 0;
  _videoBaseDTS = 0;
  _audioBaseDTS = 0;
  _realVideoDTS = 0;
  
  _chunkDuration = 0;
  _interlaced = true;
  _vfrFrame = false;
}


////////////////////////////////////////
FFMPEGDecoder::~FFMPEGDecoder() {

  avpicture_free(&_vDeinterFrame);
  //avpicture_free(&_vRGBFrame);

  if (_aFrame != NULL) {
    free(_aFrame);
    _aFrame = NULL;
  }

  if (_avFormatPrm != NULL) {
    free(_avFormatPrm);
    _avFormatPrm = NULL;
  }

  if (_audioMemory != NULL) {
    free(_audioMemory);
    _audioMemory = NULL;
  }

  if (_aResampleFrame != NULL) {
    free(_aResampleFrame);
    _aResampleFrame = NULL;
  }
}

/**********************************************************************
PUBLIC METHOD
**********************************************************************/
bool FFMPEGDecoder::open(esp_ffmpeg_params_t *p) {
  _params = p;
  return open(p->inputfile);
}

bool FFMPEGDecoder::open(char *inputfile) {
  
  avcodec_register_all();
  av_register_all();
  
  if(av_open_input_file(&_avFormatCtx,
                        inputfile,
                        NULL,
                        0,
                        _avFormatPrm) != 0) {
    
    fprintf(stderr, "Couldn't open input file: %s\n", inputfile);
    return false;
  }
  
  if(av_find_stream_info(_avFormatCtx) < 0) {
    return false; // Couldn't find stream information
  }
  
  // Dump information
  dump_format(_avFormatCtx, 0, inputfile, false);
  
  // mount streams
  _videoStream = CODEC_TYPE_UNKNOWN;
  _audioStream = CODEC_TYPE_UNKNOWN;
  for(int i=0; i< _avFormatCtx->nb_streams; i++) {
    
    int type = _avFormatCtx->streams[i]->codec->codec_type;
    
    if (type == CODEC_TYPE_VIDEO) {
      _videoStream = i;
    } else if (type == CODEC_TYPE_AUDIO) {
      
      //_aCodecCtx->channels
      if (_avFormatCtx->streams[i]->codec->channels > 2) {
        continue;
      }
      _audioStream = i;
      
    }
  }
  
  if (_videoStream != CODEC_TYPE_UNKNOWN) {
    _vCodecCtx = _avFormatCtx->streams[_videoStream]->codec;
    
    _vCodec = avcodec_find_decoder(_vCodecCtx->codec_id);
    if (_vCodec != NULL) {
      _hasVideo = true;
    } else {
      _hasVideo = false;
    }
    
    _aspectX = _vCodecCtx->sample_aspect_ratio.num;
    _aspectY = _vCodecCtx->sample_aspect_ratio.den;
    
    fprintf(stderr,"width x height: %d x %d\n",
            _vCodecCtx->width,
            _vCodecCtx->height);
    fprintf(stderr,"asX x asY: %d x %d\n",
            _aspectX, _aspectY);
    fprintf(stderr,"fps: den %d / num %d\n",
            _vCodecCtx->time_base.den,
            _vCodecCtx->time_base.num);
    fprintf(stderr,"ffmpeg: video dts %lld cur dts %lld\n",
            _avFormatCtx->streams[_videoStream]->duration,
            _avFormatCtx->streams[_videoStream]->cur_dts);
    fprintf(stderr,"ffmpeg: video start time %lld\n",
            _avFormatCtx->streams[_videoStream]->start_time);    
  } else {
    fprintf(stderr, "ffmpeg: video codec is unknown\n");
  }

  
  if (_audioStream != CODEC_TYPE_UNKNOWN) {
    _aCodecCtx = _avFormatCtx->streams[_audioStream]->codec;
    
    _aCodec = avcodec_find_decoder(_aCodecCtx->codec_id);

    fprintf(stderr,"ffmpeg: audio dts %lld\n",
            _avFormatCtx->streams[_audioStream]->duration);
    fprintf(stderr,"ffmpeg: audio start time %lld\n",
            _avFormatCtx->streams[_audioStream]->start_time);    
    
    if (_aCodec != NULL) {
      _hasAudio = true;
    } else {
      _hasAudio = false;
    }
    
    // set up audio resample ctx
    _aResampleCtx = audio_resample_init(2, 2,
                                        48000, 48000);  
  } else {
    fprintf(stderr, "ffmpeg: audio codec is unknown\n");
    return false;
  }
  
  _inputfileOpened = true;
  fprintf(stderr, "ffmpeg: parsing media OK\n");
  
  return true;
}


////////////////////////////////////////
bool FFMPEGDecoder::init(int64_t timestamp) {

  init();

  /*
    AVSEEK_FLAG_BACKWARD 1 ///< seek backward
    AVSEEK_FLAG_BYTE     2 ///< seeking based on position in bytes
    AVSEEK_FLAG_ANY      4 ///< seek to any frame, even non keyframes
  */

  //av_seek_frame(_avFormatCtx, _videoStream, timestamp * 1024 * 1024, 2);
  av_seek_frame(_avFormatCtx, _videoStream, timestamp * 1024 * 1024, 2);

  return true;
  
}

  
////////////////////////////////////////
bool FFMPEGDecoder::init() {
  if (_videoStream != CODEC_TYPE_UNKNOWN) {
    if(avcodec_open(_vCodecCtx, _vCodec) < 0) {
      return false;
    }

    _vFrame        = avcodec_alloc_frame();
    avcodec_default_get_buffer(_vCodecCtx, _vFrame);
    
    avpicture_alloc(&_vDeinterFrame, PIX_FMT_YUV420P,
                    _vCodecCtx->width, _vCodecCtx->height);
					
	//_vCodecCtx->thread_count = 4;
  }

  if (_audioStream != CODEC_TYPE_UNKNOWN) {
    if(avcodec_open(_aCodecCtx, _aCodec) < 0) {
      fprintf(stderr, "ffmpeg: err - opening audio codec\n");
      return false;
    }
	
	//_aCodecCtx->thread_count = 4;
  }
  
  return true;
}


////////////////////////////////////////
bool FFMPEGDecoder::close() {
  

  if (_videoStream != CODEC_TYPE_UNKNOWN) {

    fprintf(stderr,"ffmpeg: done - video duration %lld cur dts %lld\n",
            _avFormatCtx->streams[_videoStream]->duration,
            _avFormatCtx->streams[_videoStream]->cur_dts-
            _avFormatCtx->streams[_videoStream]->start_time);
    
    
    if (_vFrame != NULL) {
      avcodec_default_release_buffer(_vCodecCtx, _vFrame);
      av_free(_vFrame);
      _vFrame = NULL;
    }
    
    if (_vCodecCtx != NULL) {
      avcodec_close(_vCodecCtx);
      _vCodecCtx = NULL;
    }
    
    if (_vResampleCtx != NULL) {
      img_resample_close(_vResampleCtx);
      _vResampleCtx = NULL;
    }
    
  }

  if (_audioStream != CODEC_TYPE_UNKNOWN) {

    fprintf(stderr,"ffmpeg: done - audio duration %lld cur dts %lld\n",
            _avFormatCtx->streams[_audioStream]->duration,
            _avFormatCtx->streams[_audioStream]->cur_dts-
            _avFormatCtx->streams[_audioStream]->start_time);
    
    if (_aCodecCtx != NULL) {
      avcodec_close(_aCodecCtx);
      _aCodecCtx = NULL;
    }
  }

  if (_avFormatCtx != NULL) {
    av_close_input_file(_avFormatCtx);
    _avFormatCtx = NULL;
  }

  if (_aResampleCtx != NULL) {
    audio_resample_close(_aResampleCtx);
  }

  fprintf(stderr, "ffmpeg: v:a - %lld %lld\n", _videoDTS, _audioDTS);
  fprintf(stderr, "ffmpeg: codec closed\n");
  
  return true;
}

////////////////////////////////////////
bool FFMPEGDecoder::decode() {
  
  _doneVideo = 0;
  _doneAudio = 0;

  if (_packet.data != NULL) {
    av_free_packet(&_packet);
  }
  
  if ( _finishedReading != true && av_read_frame(_avFormatCtx, &_packet) <= -1 ) {
    _finishedReading = true;
  }
  
  if ( _finishedVideo != true && _hasVideo == true && _finishedReading == true ) {
    
    decodeVideo();
    
    if ( _doneVideo == 0 ) {
      fprintf(stderr, "ffmpeg: done video stream\n");
      _finishedVideo = true;
      
      /// FIXME
      //_finishedAudio = true;
    } else {
      fprintf(stderr, "ffmpeg: flush video stream\n");
      return true;
    }
  }
  
  if ( _finishedAudio != true && _hasAudio == true && _finishedReading == true ) {
    
    decodeAudio();
    /// FIXME
    _finishedAudio = true;
    
    if ( _doneAudio == 0 ) {
      fprintf(stderr, "ffmpeg: done audio stream\n");
      _finishedAudio = true;
    } else {
      fprintf(stderr, "ffmpeg: flush audio stream\n");
      return true;
    }
  }
  
  if ( _finishedReading == true && 
       (_finishedVideo == true || _hasVideo == false) && 
       (_finishedAudio == true || _hasAudio == false)) {
    
    _finished = true;
    return true;
  }
  
  /*
    if ( _finishedReading != true ) {
    
    if(av_read_frame(_avFormatCtx, &_packet) <= -1) {
    _finished = true;
    _finishedReading = true;
    //fprintf(stderr, "remain video ----->%d\n", _remainVideoBytes);
    _remainVideoBytes  = 0;
    _rawVideoBuffer    = NULL;
    decodeVideo();
    return true;
    
    if ( _finishedVideo != true) {
    
    decodeVideo();
    
    if ( _decodedVideoBytes == 0 ) {
    _finishedVideo = true;
    }
    }
    
    }
    }
  */
  
  /*
    if(av_read_frame(_avFormatCtx, &_packet) <= -1) {
    _finished = true;
    _finishedReading = true;
    //fprintf(stderr, "remain video ----->%d\n", _remainVideoBytes);
    _remainVideoBytes  = 0;
    _rawVideoBuffer    = NULL;
    decodeVideo();
    return true;
    }
  */
  
  if (_packet.stream_index == _videoStream) {
    ///// VIDEO SEQUENCE
    fprintf(stderr, "packet: video stream duration %d\n", _packet.duration);
    _videoDecodedFrameCNT++;
    
    // detect changing video farame length: interlace -> progressive
    if ((_preVideoFrameDuration - _packet.duration) < 0) {
      _changedVideoFrameDuration = true;
    }

    if (_videoBaseDTS == 0) {
      _videoBaseDTS = _packet.dts;
    }
    
    _preVideoFrameDuration = _packet.duration;
    
    detectProgressive(_packet.duration);
    
    putVideoDuration(_avFormatCtx->streams[_videoStream]->cur_dts - _packet.dts);
    
    putVideoDTS(_packet.dts);
    _pts = _packet.dts;

    //int64_t dif = _avFormatCtx->streams[_videoStream]->cur_dts - _packet.dts;
    //fprintf(stderr, "ffmpeg: cur dts %lld\n", _avFormatCtx->streams[_videoStream]->cur_dts);
    //fprintf(stderr, "ffmpeg: cur dts %lld\n", _packet.dts);
    
    //fprintf(stderr, "ffmpeg: dts diff %lld\n", _packet.dts - dtss);
    
    _totalDecodedBytes += _packet.size;
    _remainVideoBytes += _packet.size;
    _rawVideoBuffer    = _packet.data;

    //fprintf(stderr, "ffmpeg: cur dts %lld\n",
    //_avFormatCtx->streams[_videoStream]->cur_dts);
    
    decodeVideo();
    /////
  } else if (_packet.stream_index == _audioStream) {
    _totalDecodedBytes += _packet.size;
    ///// AUDIO SEQUENCE
    //fprintf(stderr, "packet: audio stream %d - %d\n", _packet.size, _packet.duration);

    if (_audioBaseDTS == 0) {
      _audioBaseDTS = _packet.dts;
    }
    
    fprintf(stderr, "packet: audio stream %lld\n", _avFormatCtx->streams[_audioStream]->cur_dts - _packet.dts);
    putAudioDuration(_avFormatCtx->streams[_audioStream]->cur_dts - _packet.dts);
    
    _remainAudioBytes += _packet.size;
    _rawAudioBuffer    = _packet.data;

    //putVideoDuration(_avFormatCtx->streams[_audioStream]->cur_dts - _packet.dts);
    
    decodeAudio();
    /////
  } else {
    fprintf(stderr, "ffmpeg: other channel\n");
  }
  
  return true;
}
  
////////////////////////////////////////
bool FFMPEGDecoder::decodeVideo() {
  
  _decodedVideoBytes = avcodec_decode_video(_vCodecCtx,
                                            _vFrame,
                                            &_doneVideo,
                                            _rawVideoBuffer,
                                            _remainVideoBytes);
  
  if (_decodedVideoBytes <= 0) {
    fprintf(stderr, "ffmpeg: failed decoding video %d\n", _doneVideo);
    return false;
  }

  _remainVideoBytes -= _decodedVideoBytes;
  //_rawVideoBuffer += _decodedVideoBytes;
  
  if (_doneVideo != 0) {
    return true;
  }
  
  return false;
}

////////////////////////////////////////
bool FFMPEGDecoder::decodeOnlyVideo() {
  int val;
  
  for (;;) {
    while (_remainVideoBytes > 0) {
      _decodedVideoBytes = avcodec_decode_video(_vCodecCtx,
                                                _vFrame,
                                                &_doneVideo,
                                                _rawVideoBuffer,
                                                _remainVideoBytes);
      
      if (_decodedVideoBytes < 0) {
        fprintf(stderr, "false\n");
        return false;
      }

      /*
      fprintf(stderr, "decoded %d\n", _decodedVideoBytes);
      fprintf(stderr, "packet size %d\n", _vPacket.size);
      */

      _remainVideoBytes -= _decodedVideoBytes;
      _rawVideoBuffer += _decodedVideoBytes;
      
      if (_doneVideo != 0) {
        //fprintf(stderr, "done %d\n", _doneVideo);
        return true;
      }
    }

    do {
      if (_vPacket.data != NULL) {
        av_free_packet(&_vPacket);
      }
      
      //val = av_read_packet(_avFormatCtx, &_vPacket);
      val = av_read_frame(_avFormatCtx, &_vPacket);
      
      if (val <= -1) {
        _finished = true;
        return true;
      }

    } while (_vPacket.stream_index != _videoStream);

    _remainVideoBytes += _vPacket.size;
    _rawVideoBuffer = _vPacket.data;
  }

  fprintf(stderr, "return\n\n\n");
  return true;
}

////////////////////////////////////////
bool FFMPEGDecoder::decodeAudio() {

  _decodedAudioBytes = avcodec_decode_audio(_aCodecCtx,
                                            (short *)_aFrame,
                                            &_doneAudio,
                                            _rawAudioBuffer,
                                            _remainAudioBytes);
  
        /*
        short* ho = (short*)_aFrame;
        static int j = 0; 
        fprintf(stderr, "block:%d\n", j++);
        for (int i = 0; i < 100; i++) {
                fprintf(stderr, "%3d:%d\n", i, ho[i]);
        }
        fprintf(stderr, "\n");
        */
         
         /*
         for (int i = 0;i < _doneAudio; i+=2) {
                 _aFrame[i/2] = _aFrame[i];
         }
         _doneAudio /= 2;
         */
                                                
        /*
         for (int i = 1005; i > 0 && _aFrame[i] == 0; i--) {
                 fprintf(stderr, "ffmpeg: data: %d pos: %d\n", _aFrame[i], i);
         }
         fprintf(stderr, "ffmpeg: bytes %d duration %d\n", _decodedAudioBytes, _doneAudio);                                             
         */

        /*
        int v, y;
        for (int i = 0; i < _doneAudio/2; i++) {
                v = _aFrame[i] & 0xFF00FF00;
                v >>= 16;
                
                y = _aFrame[i] & 0x00FF00FF;
                y <<= 16;
                
                y |= v;
                
                _aFrame[i] = y;
        }
        */


  if (_decodedAudioBytes < 0) {
    fprintf(stderr, "failed decoding audio\n");
    return false;
  }
  
  _remainAudioBytes -= _decodedAudioBytes;
  //_rawAudioBuffer += _decodedAudioBytes;


  //fprintf(stderr, "decoded %5d: remain %5d: done %5d\n", _decodedAudioBytes, _remainAudioBytes, _doneAudio);

  if (_doneAudio != 0) {
    return true;
  }

  return false;
}

////////////////////////////////////////
bool FFMPEGDecoder::decodeOnlyAudio() {
  int val;
  
  for (;;) {
    while (_remainAudioBytes > 700) {
      _decodedAudioBytes = avcodec_decode_audio(_aCodecCtx,
                                                (short *)_aFrame,
                                                &_doneAudio,
                                                _rawAudioBuffer,
                                                _remainAudioBytes);
      
      if (_decodedAudioBytes < 0) {
        fprintf(stderr, "false\n");
        return false;
      }

      //FIXME: need memcpy:CRITICAL 
      _remainAudioBytes -= _decodedAudioBytes;
      _rawAudioBuffer += _decodedAudioBytes;

      //fprintf(stderr, "decoded %5d: remain %5d: done %5d\n",
      //_decodedAudioBytes, _remainAudioBytes, _doneAudio);
      
      if (_doneAudio != 0) {
        return true;
      }
    }

    do {
      if (_aPacket.data != NULL) {
        //av_free_packet(&_aPacket);
      }

      ///val = av_read_packet(_avFormatCtx, &_aPacket);
      val = av_read_frame(_avFormatCtx, &_aPacket);
      
      if (val <= -1) {
        _finished = true;
        return true;
      }
    } while (_aPacket.stream_index != _audioStream);

    //fprintf(stderr, "read packet size %d\n", _aPacket.size);

    // Emulate FIFO Buffering
    memcpy(_audioMemory,
           _rawAudioBuffer,
           _remainAudioBytes);
    _rawAudioBuffer = _audioMemory;
    memcpy(_rawAudioBuffer + _remainAudioBytes,
           _aPacket.data,
           _aPacket.size);

    _remainAudioBytes += _aPacket.size;
  }

  return false;
}


/************************************************************
 * BOOLEAN
 ************************************************************/

////////////////////////////////////////
bool FFMPEGDecoder::hasVideo() {
  return _hasVideo;
}


////////////////////////////////////////
bool FFMPEGDecoder::hasAudio() {
  return _hasAudio;
}


////////////////////////////////////////
int FFMPEGDecoder::isDoneVideo() {
  return _doneVideo;
}


////////////////////////////////////////
int FFMPEGDecoder::isDoneAudio() {
  return _doneAudio;
}


////////////////////////////////////////
bool FFMPEGDecoder::isFinished() {
  return _finished;
}


/************************************************************
 * GET
 ************************************************************/

////////////////////////////////////////
AVPicture *FFMPEGDecoder::getVideoYUVRef() {
  return (AVPicture *)_vFrame;
}


////////////////////////////////////////
AVPicture *FFMPEGDecoder::getDeinterlacedVideoYUVRef() {
  
  avpicture_deinterlace(&_vDeinterFrame, (AVPicture*)_vFrame,
                        PIX_FMT_YUV420P, _vCodecCtx->width, _vCodecCtx->height);
  return &_vDeinterFrame;
}

////////////////////////////////////////
uint8_t *FFMPEGDecoder::getVideoYRef() {
  AVPicture* p =  (AVPicture *)_vFrame;

  return p->data[0];
}


////////////////////////////////////////
int FFMPEGDecoder::getVideoYSize() {
  return ((AVPicture *)_vFrame)->linesize[0];
}


////////////////////////////////////////
uint8_t *FFMPEGDecoder::getVideoURef() {
  return ((AVPicture *)_vFrame)->data[1];
}


////////////////////////////////////////
int FFMPEGDecoder::getVideoUSize() {
  return ((AVPicture *)_vFrame)->linesize[1];
}


////////////////////////////////////////
uint8_t *FFMPEGDecoder::getVideoVRef() {
  return ((AVPicture *)_vFrame)->data[2];
}


////////////////////////////////////////
int FFMPEGDecoder::getVideoVSize() {
  return ((AVPicture *)_vFrame)->linesize[2];
}


////////////////////////////////////////
int64_t FFMPEGDecoder::getFileSize() {
  return _avFormatCtx->file_size;
}


////////////////////////////////////////
int64_t FFMPEGDecoder::getBytePosition() {
  return _totalDecodedBytes;
}


////////////////////////////////////////
int64_t FFMPEGDecoder::getPts() {
  return _packet.pts;
}


////////////////////////////////////////
int64_t FFMPEGDecoder::getDts() {
  return _packet.dts;
}


////////////////////////////////////////
int32_t *FFMPEGDecoder::getAudioRAWRef() {
  if (_hasAudio == true) {
    return _aFrame;
  }

  return NULL;
}


////////////////////////////////////////
int FFMPEGDecoder::getAudioSampleRate() {
  if (_hasAudio == true) {
    return _aCodecCtx->sample_rate;
  } else {
    return 0;
  }
}


////////////////////////////////////////
int FFMPEGDecoder::getAudioChannels() {
  if (_hasAudio == true) {
    return _aCodecCtx->channels;
  } else {
    return 0;
  }
}

////////////////////////////////////////
int FFMPEGDecoder::getDecodedAudioBytes() {
  if (_hasAudio == true) {
    return _decodedAudioBytes;
  }

  return 0;
}


////////////////////////////////////////
int FFMPEGDecoder::getTotalDecodedAudioBytes() {
  return _totalDecodedAudioBytes;
}


////////////////////////////////////////
int FFMPEGDecoder::getWidth() {
  if (_hasVideo == true) {
    return _vCodecCtx->width;
  }
  return false;
}


////////////////////////////////////////
int FFMPEGDecoder::getHeight() {
  if (_hasVideo == true) {
    return _vCodecCtx->height;
  }
  return false;
}


////////////////////////////////////////
int FFMPEGDecoder::getFpsNum() {
  if (_hasVideo == true) {
    return _vCodecCtx->time_base.num;
  }
  return false;
}


////////////////////////////////////////
int FFMPEGDecoder::getFpsDen() {
  if (_hasVideo == true) {
    return _vCodecCtx->time_base.den;
  }
  return false;
}

////////////////////////////////////////
int FFMPEGDecoder::getAspectX() {
        return _aspectX;
}

////////////////////////////////////////
int FFMPEGDecoder::getAspectY() {
        return _aspectY;
}


////////////////////////////////////////
int FFMPEGDecoder::getFrameDuration() {
  return _packet.duration;
}


////////////////////////////////////////
int FFMPEGDecoder::getVideoFrameDuration() {
  return getNextVideoDuration();
  //return getNextVideoDTS();
  
  //return _packet.duration;
}


////////////////////////////////////////
int64_t FFMPEGDecoder::getDecodedVideoFrameCNT() {
  return _videoDecodedFrameCNT;
}


////////////////////////////////////////
bool FFMPEGDecoder::didChangeVideoFrameDuration() {
        return _changedVideoFrameDuration;
}


////////////////////////////////////////
int FFMPEGDecoder::getDecodedAudioSampleNum() {
  if (_hasAudio == true) {
    return _doneAudio;
  }

  return 0;
}


////////////////////////////////////////
AVPicture* FFMPEGDecoder::getResampledVideoYUVRef() {

  // FIXME: input maybe a deinterlaced frame
  img_resample(_vResampleCtx,
               &_vResampleFrame, 
               (const AVPicture *)_vFrame);
  
  return &_vResampleFrame;
}


////////////////////////////////////////
AVPicture* FFMPEGDecoder::getResampledVideoYUVRef(AVPicture *in) {
  // FIXME: input maybe a deinterlaced frame
  img_resample(_vResampleCtx,
               &_vResampleFrame, 
               (const AVPicture *)in);
  
  return &_vResampleFrame;
}


////////////////////////////////////////
AVPicture* FFMPEGDecoder::convertYUVtoRGB(AVPicture *in) {
  static int a = 0;

  if (a == 0) {
    avpicture_alloc(&_vRGBFrame, PIX_FMT_RGB24,
                    _vCodecCtx->width, _vCodecCtx->height);
    a = 1;
  }
  
  img_convert(&_vRGBFrame, PIX_FMT_RGB24,
              in, PIX_FMT_YUV420P,
              _vCodecCtx->width, _vCodecCtx->height);
  
  return &_vRGBFrame;
  
}


////////////////////////////////////////
void FFMPEGDecoder::convertRGBtoYUV(AVPicture *in, AVPicture *out) {
  img_convert(out, PIX_FMT_YUV420P,
              in,  PIX_FMT_RGB24,
              _vCodecCtx->width, _vCodecCtx->height);
}


////////////////////////////////////////
int FFMPEGDecoder::setVideoResampleContext(int width, int height) {
        
  _vResampleCtx = img_resample_init(width, 
                                    height,
                                    getWidth(),
                                    getHeight());
  
  if (_vResampleCtx == NULL) {
    return -1;
  }
  
  avpicture_alloc(&_vResampleFrame,
                  PIX_FMT_YUV420P,
                  width, height);
  
  return 0;
}


void FFMPEGDecoder::seekToTop() {
  if (_hasAudio == true) {
    av_seek_frame(_avFormatCtx, _audioStream, 0, AVSEEK_FLAG_BYTE);
  }
  
  if (_hasVideo == true) {
    av_seek_frame(_avFormatCtx, _videoStream, 0, AVSEEK_FLAG_BYTE);
  }
}


////////////////////////////////////////
void FFMPEGDecoder::putVideoDuration(int value) {

  //static int off = 0;
  static int64_t preVideoDTS = 0;
  static int64_t dtsOffset = 0;
    
  // local val duration keeps result duration
  int duration = value;
  
  /*
  static bool firstFrame = true;
  if (firstFrame == true) {
    int d = _avFormatCtx->streams[_videoStream]->start_time - _avFormatCtx->streams[_audioStream]->start_time;
    d = (abs(d) < value) ? d : 0;

    //value += d;

    
    firstFrame = false;

    fprintf(stderr, "ffmpeg: first frame %d - %d\n", value, d);
  } else {
    duration = value;
  }
  */

  // range check
  if (duration > 4600 || duration < 2900) {
    
    fprintf(stderr, "ffmpeg: error. the frame duration is out of range.\n");
    
    if (_vFrame != NULL) { // vfr
      
      if (_vFrame->interlaced_frame == false) {
        duration = 3754;
        value = 3754;
      } else {
        duration = 3003;
        value = 3003;
      }
    } else { // cfr
      duration = 3003;
    }
    
  }
  
  //int64_t a = _avFormatCtx->streams[_audioStream]->cur_dts;
  //int t = v - a ;
  //int s = _videoDTS - _audioDTS;
  //int u = abs(t-s) - off;

  int64_t v = _avFormatCtx->streams[_videoStream]->cur_dts;
  //int64_t p = _packet.dts + value;
  int64_t lx = _realVideoDTS - _videoDTS;
  int ly = 0;

  if (preVideoDTS > v) {
    dtsOffset += preVideoDTS;
  }
  preVideoDTS = v;
  
  if (lx > 0) {
    ly = 20;
  } else {
    lx = -1*lx;
    ly = -20;
  }

  //fprintf(stderr, "ffmpeg: cur %lld packet %lld - out %lld - real %lld - diff %lld\n", v, p, _videoDTS, _realVideoDTS, _realVideoDTS - _videoDTS);
  fprintf(stderr, "ffmpeg: cur %lld - out %lld - real %lld - diff %lld\n", v, _videoDTS, _realVideoDTS, _realVideoDTS - _videoDTS);

  if (_vfrFrame == true) {
    
    duration = 3754;
    
    if (lx > 780 && _audioDTS != 0) {
      duration += ly;
    }
    
    fprintf(stderr, "ffmpeg: base - %lld diff - %lld offset %lld\n", _videoBaseDTS, (v+dtsOffset) -_videoDTS, dtsOffset);
    if (( (v + dtsOffset) - _videoDTS) < _videoBaseDTS) {
      fprintf(stderr, "ffmpeg: using base correction\n");
      duration -= 20;
      value -= 20;
    }
    
  }
  
  _videoDTS += duration;
  _realVideoDTS += value;
  
  _dRing[_dRingMemory] = duration;
  _dRingMemory++;
  _dRingMemory = ((_dRingMemory % _dRingSize) == 0) ? 0 : _dRingMemory;
}

int64_t FFMPEGDecoder::getNextVideoDuration() {
  int64_t ret = _dRing[_dRingData];
  
  _dRingData++;
  
  _dRingData = ((_dRingData % _dRingSize) == 0) ? 0 : _dRingData;
  
  return ret;
}

void FFMPEGDecoder::putAudioDuration(int64_t duration) {

  if (duration < 0 || duration > 6800) {
    duration = 0;
  }
  
  _audioDTS += duration;
}



void FFMPEGDecoder::putVideoDTS(int64_t duration) {

  _dtsRing[_dtsRingMemory] = duration;

  _dtsRingMemory++;
  
  _dtsRingMemory = ((_dtsRingMemory % _dtsRingSize) == 0) ? 0 : _dtsRingMemory;
}




int64_t FFMPEGDecoder::getNextVideoDTS() {
  int64_t ret = _dtsRing[_dtsRingData];
  
  _dtsRingData++;
  
  _dtsRingData = ((_dtsRingData % _dtsRingSize) == 0) ? 0 : _dtsRingData;
  
  return ret;
}


bool FFMPEGDecoder::detectProgressive(int d) {
  static int num = 0;
  
  num++;
  _chunkDuration += d;

  _interlaced = _vFrame->interlaced_frame;
  
  if ( num == 6 ) {
    if ( _chunkDuration <= 18018 ) {
      //_interlaced = true;
      _vfrFrame = false;
    } else {
      //_interlaced = false;
      _vfrFrame = true;
    }
    
    num = 0;
    _chunkDuration = 0;
  }
  
  if (d == 4505) {
    //_interlaced = false; // i.e progressive
    _vfrFrame = true;
  }
  
  return true;
}


bool FFMPEGDecoder::isInterlaced() {
  if (_vFrame->interlaced_frame == 1) {
    return true;
  }

  return false;
}

int64_t FFMPEGDecoder::getCurrentDTS() {
  return  _avFormatCtx->streams[_videoStream]->cur_dts;
}

bool FFMPEGDecoder::isVFRFrame() {
  return _vfrFrame;
}

// value range 0 to 1000
// based on dts infomation
int FFMPEGDecoder::getProgress() {

  AVStream *vst = NULL;
  int ret = 0;

  if ( _hasVideo == true ) {
    vst = _avFormatCtx->streams[_videoStream];
    ret = (int)((_packet.dts * 1000) / (vst->duration + vst->start_time));
  } else if ( _hasAudio == true) {
    vst = _avFormatCtx->streams[_audioStream];
    ret = (int)((_packet.dts * 1000) / (vst->duration + vst->start_time));
  }

  return ret;
}

/**********************************************************************
TEST FUNCTION
**********************************************************************/
#ifdef TEST_FFMPEG
int main(int argc, char *argv[]) {

  AVFrame *v;
  AVFrame *rgb = avcodec_alloc_frame();
  int numBytes = avpicture_get_size(PIX_FMT_RGB24, 740, 480);
  FFMPEGDecoder *d = new FFMPEGDecoder();
  uint8_t      *buf;
  char ppmfilename[16];
  FILE *ppm;

  buf = new uint8_t[numBytes];
  avpicture_fill((AVPicture *)rgb, buf, PIX_FMT_RGB24, 740, 480);

  if (d->init(argv[1]) == false) {
    puts("error occured");
  } else {
    puts("goes well");
  }
  d->open();
  
  for (int i = 0; i < 2000; i++) {
    fprintf(stderr, "decoded %3d frames\n", i+1);
    d->decode();
    v = d->getVideoYUVRef();


    img_convert((AVPicture *)rgb, PIX_FMT_RGB24, (AVPicture *)v,
                PIX_FMT_YUV420P, 720, 480);

    if (i < 0) {
      sprintf(ppmfilename, "ppm-%d", i);
      ppm = fopen(ppmfilename, "wb");
      fprintf(ppm, "P6 %d %d 255\n", 720, 480);
      for(int j = 0; j < 480; j++) {
        fwrite(rgb->data[0]+j*rgb->linesize[0], 1, 720 * 3, ppm);
      }
      fclose(ppm);
    }
    
  }

  av_free(rgb);
  d->close();
  delete[] buf;
  delete d;

  return 0;
}
#endif
