/********************************************************************************
Copyright (C) Takuya.S

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 ********************************************************************************/

#ifndef _MP4HANDLER_H_
#define _MP4HANDLER_H_

#include <stdio.h>
#include <stdlib.h>

#ifdef _USE_MAKE_
#include <mp4.h>
#else
#include "mp4.h"
#endif // _USE_MAKE_

#define MAX_MP4_FILENAME 128

class MP4Handler {

 private:
  MP4FileHandle _hFile;
  MP4TrackId _idVideoTrack;
  MP4TrackId _idAudioTrack;
  char *_filename;
  int _timeScale;
  int _width;
  int _height;
  
  int _vDelay;
  int _aDelay;

  int _vFrameCount;
  int _aFrameCount;

  long unsigned int _ascLength;
  unsigned char *_asc;

  bool _optimize;
  
  bool _hasVideoTrack;
  bool _hasAudioTrack;

  int64_t _videoPTS;
  int64_t _audioPTS;

  bool optimizeMP4File();

 public:
  MP4Handler();
  ~MP4Handler();

  bool init();
  bool open(const char *);
  bool open();
  bool close();

  bool setMP4TimeScale(int scale);

  // Video
  bool addH264VideoTrack(u_int8_t *, int32_t w, int32_t h, int32_t d);
  bool setH264SequenceParameter(u_int8_t *, int);
  bool setH264PictureParameter(u_int8_t *, int);
  bool writeMP4VideoSample(u_int8_t *, int, u_int64_t);
  bool writeMP4VideoSample(u_int8_t *, int, u_int64_t, bool);
  bool writeMP4VideoSample(u_int8_t *, int, u_int64_t, int, bool);
  
  bool setVideoDelayTime(int);

  // Audio
  bool addAACAudioTrack(u_int32_t);
  bool addAACAudioTrack(u_int32_t, uint8_t*, long unsigned int);
  bool setMP4AudioProfileLevel();
  bool writeMP4AudioSample(u_int8_t *, int, u_int64_t);

  bool setAudioDelayTime(int);

  unsigned char ** getASCRef();
  long unsigned int *getASCLengthRef();

  void setWidth(int);
  void setHeight(int);  

};

#endif
