/********************************************************************************
Copyright (C) Takuya.S

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
********************************************************************************/
#ifndef _X264ENCODER_H_
#define _X264ENCODER_H_

#ifdef WIN32
#include "MP4handler.h"
#endif

#ifdef _USE_MAKE_
#include <x264.h>
#else
#include "x264.h"
#endif

typedef struct {

  int pass;

  int width;
  int height;
  
  int sar_width;
  int sar_height;
  
  int fps_den;
  int fps_num;

  int e_mode;
#define X264_QMODE_CRF 0
#define X264_QMODE_CBR 1
#define X264_QMODE_CQP 2
  
  int crf;
  int cbr;
  int cqp;
  int qp_cons;  
  int qp_min;
  int qp_max;
  int qp_step;
  
  int nr;
  
  int threads;  
  int subpel_refine;
  int frame_reference;  
  int me_method;
  int bframe;
  int df_alpha;
  int df_beta;

  int bframe_adaptive;
  int psnr;
  int weighted_bipred;
  int bidir_me;
  int fast_pskip;
  int transform_8x8;
  
  int key_max;
  
  char* log_file;

  bool resizedPicture;

} esp_x264_params_t;

class X264Encoder {
private:
  x264_t         *_hEncoder;
  x264_nal_t     *_nal;
  int            _nalCount;
  x264_picture_t _picIn;
  x264_picture_t _picOut;
  x264_param_t   *_param;
  int64_t        _bytesWritten;
  int64_t        _totalBytesWritten;
  int            _nalEncodeBufferSize;
  u_int8_t       *_nalEncodeBuffer;
  int _width;
  int _height;
  int _num;
  int _den;
  u_int8_t *_sps;
  u_int8_t *_pps;
  int _spsLen;
  int _ppsLen;
  bool _changeSPS;
  bool _changePPS;
  bool _headerChanged;
  
  //esp_x264_params_t *param;

  bool encodeNal();
  void setSPS(int );
  void setPPS(int );
  
  void setDefaultParam();
  void setTotalBytesWrittenZero(); // called befor encode

public:
  X264Encoder();
  ~X264Encoder();
  
  bool init(esp_x264_params_t *);
  bool init();
  bool open(int32_t w, int32_t h, int32_t d, int32_t n);
  bool open();
  
  bool close();
  bool encode();
  bool encode(int );
  bool flush();

  x264_picture_t *getX264PictureRef();
  u_int8_t *getNalEncodedBufferRef();
  int     getBytesWritten();
  int64_t getTotalBytesWritten();
  int getNalCount();

  void setWidth(int);
  void setHeight(int);
  void setFrameRateFactor(int num, int den);

  int getWidth();
  int getHeight();

  u_int8_t *getSPSRef();
  u_int8_t *getPPSRef();
  int getSPSLen();
  int getPPSLen();
  bool isChangedHeader();

  u_int64_t getPTS();
  u_int8_t getSliceType();
};

#endif // _X264ENCODER_H_
